/*************************************************************************
* fwSDLApp - Application class for SDL
* Author: Thomas Seban
* Date:   17.06.2006
*
* Simple class for a sdl main loop
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string>

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_mixer.h>
#include <SDL/SDL_ttf.h>

#include "fwSDLLog.h"
#include "fwSDLApp.h"
#include "fwGlobals.h"

#ifdef GP2X
#include "gp2x.h"
#endif

static SDLApp* Application = NULL;

SDLApp* GetApplication() {
    return Application;
}


void SDLAppTerminate(void)
{
	SDL_Quit();
#ifndef DEBUG
#ifdef GP2X
	// XXXX chdir("/usr/gp2x");
	// XXXX execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
#endif
#endif
}

#ifdef GP2X
// Converts gp2x button states to keyboad states
void SDLApp::gp2xConvertKey(Uint8 *key_table) {

    // joystick
    key_table[SDLK_LEFT]   = 0;
    key_table[SDLK_RIGHT]  = 0;
    key_table[SDLK_UP]     = 0;
    key_table[SDLK_DOWN]   = 0;

    key_table[SDLK_a]      = 0;
    key_table[SDLK_b]      = 0;
    key_table[SDLK_x]      = 0;
    key_table[SDLK_y]      = 0;

    key_table[SDLK_l]      = 0;
    key_table[SDLK_r]      = 0;

    key_table[SDLK_s]      = 0;

    key_table[SDLK_l]      = 0;
    key_table[SDLK_2]      = 0;

    key_table[SDLK_RETURN] = 0;
    key_table[SDLK_SPACE]  = 0;

    // top buttons
    if (gp2x_ButtonPressed(gp2x_LButton)) {
      key_table[SDLK_l] = 1;
    }

    if (gp2x_ButtonPressed(gp2x_RButton)) {
      key_table[SDLK_r] = 1;
    }

    // right buttons
    if (gp2x_ButtonPressed(gp2x_A)) {
      key_table[SDLK_a] = 1;
    }

    if (gp2x_ButtonPressed(gp2x_B)) {
      key_table[SDLK_b] = 1;
    }

    if (gp2x_ButtonPressed(gp2x_X)) {
      key_table[SDLK_x] = 1;
    }

    if (gp2x_ButtonPressed(gp2x_Y)) {
      key_table[SDLK_y] = 1;
    }

    // right buttons select and start
    if (gp2x_ButtonPressed(gp2x_Start)) {
      key_table[SDLK_RETURN] = 1;
    }

    if (gp2x_ButtonPressed(gp2x_Select)) {
      key_table[SDLK_SPACE] = 1;
    }

    // Joysticks
    if (gp2x_JoystickPressed(gp2x_JoyUp, false)) {
      key_table[SDLK_UP] = 1;
    }

    if (gp2x_JoystickPressed(gp2x_JoyDown, false)) {
      key_table[SDLK_DOWN] = 1;
    }

    if (gp2x_JoystickPressed(gp2x_JoyLeft, false)) {
      key_table[SDLK_LEFT] = 1;
    }

    if (gp2x_JoystickPressed(gp2x_JoyRight, false)) {
      key_table[SDLK_RIGHT] = 1;
    }

    if (gp2x_JoystickPressed(gp2x_Center, false)) {
      key_table[SDLK_s] = 1;
    }

    // Volume buttons
    if (gp2x_VolumePressed(gp2x_VolMinus)) {
      key_table[SDLK_1] = 1;
    }

    if (gp2x_VolumePressed(gp2x_VolPlus)) {
      key_table[SDLK_2] = 1;
    }
}
#endif

// SDLApp ----------------------------------------------------------------

void SDLApp::ClearScreen() {
    SDL_FillRect (mScreen, NULL, SDL_MapRGB (mScreen->format, 0, 0, 0));
}

void SDLApp::FlipScreen() {
    SDL_Flip(mScreen);
}

void SDLApp::Render() {

    InternalBeforeDrawScene(mScreen);
    InternalDrawScene(mScreen);
    InternalAfterDrawScene(mScreen);


    FlipScreen();

    mFrames++;
    if ((SDL_GetTicks() - mFPSTicks) >= 990) {
        mFPS      = (mFrames * 1000.0) / (SDL_GetTicks() - mFPSTicks);
        mFrames   = 0;
        mFPSTicks = SDL_GetTicks();
    }
}

SDLApp::SDLApp() {
    if (Application) {
        fprintf (stderr, "SDLApp already exists\n");
        exit(2);
    }

#ifdef GP2X
    if (!initphys())
#endif

    mSDLFlags   = SDL_INIT_VIDEO | SDL_INIT_AUDIO;
    mFPS        = 0.0;
    mFrames     = 0;
    mDone       = false;
    mUseTTFont  = false;
    mKeyStates  = NULL;
    Application = this;
}

SDLApp::~SDLApp() {
#ifdef GP2X
  mmuunhack();
  closephys();
#endif

  Application = NULL;
}

void SDLApp::InitSDL() {
    fprintf(stdout, "Init SDL...\n");

	if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0) {
        Log.Add("SDLApp::InitSDL", SDL_GetError());
		throw "InitSDL failed";
	}

	atexit (SDLAppTerminate);

    fprintf(stdout, "Init Audio...\n");
#ifdef GP2X
    // Initialize SDL_mixer for GP2X, buffer is set lower than PC
    if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, AUDIO_S16, MIX_DEFAULT_CHANNELS, 128) == -1) {
        Log.Add("SDLApp::InitSDL", SDL_GetError ());
		throw "InitSDL:Audio failed";
    }
#else
	// Initialize SDL_mixer for PC, buffer is set higher
    if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, AUDIO_S16, MIX_DEFAULT_CHANNELS, 2048) == -1) {
        Log.Add("SDLApp::InitSDL", SDL_GetError ());
		throw "InitSDL:Audio failed";
    }
#endif

    fprintf(stdout, "Gettting SDL_GetVideoInfo...\n");
    mVideoInfo = SDL_GetVideoInfo();
    if (!mVideoInfo) {
        Log.Add("SDLApp::InitSDL - SDL_GetVideoInfo", SDL_GetError());
        throw "InitSDL:Video failed";
    }

    mVideoFlags = SDL_HWPALETTE | SDL_DOUBLEBUF | SDL_HWACCEL;
#ifdef GP2X
    mVideoFlags = mVideoFlags | SDL_FULLSCREEN;
#endif
    if (mVideoInfo->hw_available == 1) {
        Log.Add("SDLApp::InitSDL", "Hardware Surface enabled...");
        mVideoFlags = mVideoFlags | SDL_HWSURFACE;
    } else {
        Log.Add("SDLApp::InitSDL", "Surface Surface enabled...");
        mVideoFlags = mVideoFlags | SDL_SWSURFACE;
    }

	/* Set 320x240 16-bits video mode */
    fprintf(stdout, "Init Video...\n");
	mScreen = SDL_SetVideoMode (320, 240, 16, mVideoFlags);
	if (mScreen == NULL) {
        Log.Add("SDLApp::InitSDL", SDL_GetError ());
		throw "InitSDL failed";
	}

    // disable mouse cursor, must be after calling SDL_SetVideoMode()
	SDL_ShowCursor(SDL_DISABLE);

    // init MMU hack by Squidge and Notaz
#ifdef GP2X
	mmuhack();
#endif

    Log.Add("SDLApp::InitSDL", "Done");
}

void SDLApp::InternalInit() {
}

void SDLApp::InternalShutdown() {
    // FontList->clear;
    // ImageList->clear;
}

void SDLApp::InternalHandleEvent(SDL_Event *event) {

    switch (event->type) {
        case SDL_KEYDOWN:
            if (KeyPressed(SDLK_ESCAPE)) mDone = 1;
            break;
        case SDL_QUIT:
            mDone = 1;
            break;
        default:
            break;
  }
}

void SDLApp::InternalBeforeDrawScene(SDL_Surface *Screen) {
}

void SDLApp::InternalDrawScene(SDL_Surface *Screen) {
}

void SDLApp::InternalAfterDrawScene(SDL_Surface *Screen) {
}

void SDLApp::SetCommandLine(int aArgCount, char* aArgs[]) {
  InitCommandLine();
}

bool SDLApp::ExistsParam(char* aParam) {
	return false;
}

char* SDLApp::GetCommandLineValue(char * aParam) {
    return NULL;
}

void SDLApp::Init() {
    InitSDL();
    InternalInit();
}

void SDLApp::Stop() {
    mDone = true;
}

void SDLApp::Run() {
    SDL_Event event;

    mDone          = false;
    mFrames        = 0;
    mFPSTicks      = SDL_GetTicks();
    mLastFrameTime = 0;
    mMaxFrameTime  = 0;

    while (!mDone) {
        Uint32 FrameTime = SDL_GetTicks();

        mKeyStates = SDL_GetKeyState(NULL);
#ifdef GP2X
        gp2xConvertKey(mKeyStates);
#endif
        if (SDL_PollEvent (&event)) {
            InternalHandleEvent(&event);
            while (SDL_PollEvent (&event)) {
                InternalHandleEvent(&event);
            }
        }

        ClearScreen();
        Render();

        // Setup a constant refresh rate
        mLastFrameTime = (SDL_GetTicks() - FrameTime);
        if (mLastFrameTime > mMaxFrameTime) {
            fprintf(stdout, "Frametime: %d\n", mLastFrameTime);
            mMaxFrameTime = mLastFrameTime;
        };

        if (mLastFrameTime < MAXFRAMETIME) {
            SDL_Delay(MAXFRAMETIME - mLastFrameTime);
        };
    }

    InternalShutdown();
    Mix_CloseAudio();
}

float SDLApp::getFPS() {
  return mFPS;
};


bool SDLApp::KeyPressed(int Key, bool aWaitUntilRelease) {
    SDL_Event event;

    if (mKeyStates) {
        if (mKeyStates[Key]) {
            while (aWaitUntilRelease) {
                // if waiting for key press, check and convert
                // key states
                while (SDL_PollEvent (&event)) {
                    InternalHandleEvent(&event);
                }
                mKeyStates = SDL_GetKeyState(NULL);
#ifdef GP2X
                gp2xConvertKey(mKeyStates);
#endif
                if (!mKeyStates[Key]) {
                    break;
                }
            }

            return true;
        }
    }

    return false;
}

Uint32 SDLApp::RGB(Uint8 r, Uint8 g, Uint8 b) {
    return SDL_MapRGB(mScreen->format, r, g, b);
}
